/*------------------------------------------------------------------------------

   Function: adjacent_bin_number

   Returns type: unsigned long

   Description:	This function returns the number of the quadsphere bin that
		is directly adjacent to the bin specified by the first argument.
		The third argument, side, should have one of the four values
		defined in quadsphere.h to indicate whether the returned
		number refers to a bin which is left of, right of, above, or
		below the bin specified by the first argument, bin_number.
		The local sense of left, right, above, and below is determined
		by the cube face that contains the input bin number.

		If any of the arguments to this function fall outside of their
		proper ranges or are inconsistent with each other, then this
		function returns a value of -1 to indicate an error condition.


   Parameters: (in calling order)
      Type	Name		I/O	Description
      ----	----		---	-----------
      long	bin_number	In	Quadsphere bin number
      int	level		In	Quadsphere bin level
					(See comments in
					compute_quadsphere_bin_number.c
					for a description of the quadsphere.)
      int	side		In	Specifies whether output bin
					should be left, right, above,
					or below the input bin  (see
					quadsphere.h)

   Modification history:
      Programmer	Date		Description of change
      ----------	----		---------------------
      Norman Kuring	17-Feb-1993	Original development

------------------------------------------------------------------------------*/

#include "quadsphere.h"

unsigned long spread_bits(unsigned short number);

long
adjacent_bin_number(bin_number,level,side)
long		bin_number;
int		level;
int		side;
{

  int		cube_face;
  int		iu,iv,in;
  int		maximum_dimension;
  long		error_value = -1;

  if(level < 1 || level > 14)
    return(error_value);

  maximum_dimension = (1<<level) - 1;

  cube_face = bin_number/(1 << 2*level);

  if(cube_face < 0 || cube_face > 5)
    return(error_value);

  bin_number -= cube_face * (1 << 2*level);

  for(iu = 0, iv = 0, in = 1; bin_number; in *= 2){

    iu += (bin_number & 1) * in;
    bin_number >>= 1;

    iv += (bin_number & 1) * in;
    bin_number >>= 1;

  }

  switch(side){
    case QS_LEFT:
      if(--iu < 0){
        switch(cube_face){
          case 0:
            cube_face = 4;
            iu = maximum_dimension - iv;
            iv = maximum_dimension;
            break;
          case 1:
            cube_face = 4;
            iu = maximum_dimension;
            break;
          case 2: case 3: case 4:
            cube_face--;
            iu = maximum_dimension;
            break;
          case 5:
            cube_face = 4;
            iu = iv;
            iv = 0;
            break;
        }
      }
      break;
    case QS_RIGHT:
      if(++iu > maximum_dimension){
        switch(cube_face){
          case 0:
            cube_face = 2;
            iu = iv;
            iv = maximum_dimension;
            break;
          case 1: case 2: case 3: case 4:
            cube_face = cube_face%4 + 1;
            iu = 0;
            break;
          case 5:
            cube_face = 2;
            iu = maximum_dimension - iv;
            iv = 0;
            break;
        }
      }
      break;
    case QS_TOP:
      if(++iv > maximum_dimension){
        switch(cube_face){
          case 0: case 3:
            cube_face = 3 - cube_face;
            iu = maximum_dimension - iu;
            iv = maximum_dimension;
            break;
          case 1:
            cube_face = 0;
            iv = 0;
            break;
          case 2:
            cube_face = 0;
            iv = iu;
            iu = maximum_dimension;
            break;
          case 4:
            cube_face = 0;
            iv = maximum_dimension - iu;
            iu = 0;
            break;
          case 5:
            cube_face = 1;
            iv = 0;
            break;
        }
      }
      break;
    case QS_BOTTOM:
      if(--iv < 0){
        switch(cube_face){
          case 0:
            cube_face = 1;
            iv = maximum_dimension;
            break;
          case 1:
            cube_face = 5;
            iv = maximum_dimension;
            break;
          case 2:
            cube_face = 5;
            iv = maximum_dimension - iu;
            iu = maximum_dimension;
            break;
          case 3: case 5:
            cube_face = 8 - cube_face;
            iu = maximum_dimension - iu;
            iv = 0;
            break;
          case 4:
            cube_face = 5;
            iv = iu;
            iu = 0;
            break;
        }
      }
      break;
    default:
      return(error_value);
  }

  return( (long)
  spread_bits(iu) + (spread_bits(iv) << 1) + cube_face * (1 << 2*level)
  );

}/*end adjacent_bin_number()*/



/*------------------------------------------------------------------------------
   Function: spread_bits

   Returns type: unsigned long

   Description: This function returns a value that corresponds to the
		input value with an extra zero-bit inserted between each
		of the input value's bits.  A full 16-bit value is therefore
		stretched out to 31 bits in length.

   Parameters: (in calling order)
      Type		Name	I/O	Description
      ----		----	---	-----------
      unsigned short	number	In	Number to be spread out bitwise

   Modification history:
      Programmer	Date		Description of change
      ----------	----		---------------------
      Norman Kuring	03-Nov-1992	Original development

------------------------------------------------------------------------------*/

unsigned long
spread_bits(number)
unsigned short  number;
{
  short		loop;
  unsigned long new_number;

  for(loop=0, new_number=0; loop<16; loop++)
    new_number += ( (number & 1<<loop) << loop);

  return(new_number);

}/*end spread_bits()*/
