#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "spatial_info.h"

#define		SUCCESS			0
#define		ADJ_BIN_NUM_ERROR	1
#define		REALLOC_ERROR		2
#define		LOCATE_BIN_ERROR	3
#define		LOGIC_ERROR		4
#define		MALLOC_ERROR		5
#define		INCONSISTENT_DATA_ERROR	6
#define		COORDINATE_ERROR	7

#define PI	3.14159265358979323846

/*------------------------------------------------------------------------------

   Function: generate_area_outlines

   Returns type: int

   Description:	This function fills out the ListOfAreas structure with a set
		of outlines and coordinate extrema for each quadsphere bin
		mask that is already stored in the structure.

		This function returns one of the status values defined above.


   Parameters: (in calling order)
      Type		Name	I/O	Description
      ----		----	---	-----------
      ListOfAreas	areas	In/Out	See spatial_info.h

   Modification history:
      Programmer	Date		Description of change
      ----------	----		---------------------
      Norman Kuring	23-Feb-1993	Original development
      Norman Kuring	11-Mar-2004	Adapt to non-X-callback use

------------------------------------------------------------------------------*/

int
generate_area_outlines(areas)
ListOfAreas		areas;
{

  int			area_index;
  int			return_status;
  Segment		*segment_array;
  int32_t		number_of_segments;

  /* Function declarations */
  int	generate_segment_array(char *mask, Segment **segarr, int32_t *numsegs);
  int	generate_linked_lists(Segment *segarr, int32_t numsegs, OutlineList *ol);
  int	find_geographical_extrema(AreaOfInterest *aoi);

  for(area_index = 0; area_index < areas.n; area_index++){

    if(areas.area[area_index].pgons.n == 0){

      if( (return_status =
      generate_segment_array(
      areas.area[area_index].bmsk,
      &segment_array,&number_of_segments)
      ) != SUCCESS)
        return(return_status);

      if( (return_status =
      generate_linked_lists(
      segment_array,number_of_segments,
      &areas.area[area_index].pgons)
      ) != SUCCESS)
        return(return_status);

      /* Deallocate memory now that we are through with it. */
      if(segment_array != NULL)
        free(segment_array);

      /* Determine the coordinates of the rectangle (in a cylindrical-equi-
       * rectangular projection) that circumscribes this area of interest. */
      if( (return_status =
      find_geographical_extrema(&areas.area[area_index])
      ) != SUCCESS)
        return(return_status);

    }

  }

  return(SUCCESS);

}/*end generate_area_outlines()*/






/*------------------------------------------------------------------------------

   Function: generate_segment_array

   Returns type: int

   Description:	This function produces an array of quadsphere bin side
		specifications that is based on the input bin mask.
		Of all the possible bin sides contained in the quadsphere,
		only those having a selected bin on just one side will be
		included in the output array.  This has the effect of
		producing a boundary around each contiguous cluster of
		selected bins.

		This function returns one of the status values defined
		at the beginning of this file.


   Parameters: (in calling order)
      Type		Name			I/O	Description
      ----		----			---	-----------
      unsigned char *	bin_mask		In	1 bit per quadsphere bin
      Segment **	segment_array		Out	Array of earth-
							coordinate pairs
      long *		number_of_segments	Out	Number of elements in
							segment_array

   Modification history:
      Programmer	Date		Description of change
      ----------	----		---------------------
      Norman Kuring	23-Feb-1993	Original development
      Norman Kuring	11-Mar-2004	Adapt to non-X-callback use

------------------------------------------------------------------------------*/

int
generate_segment_array(bin_mask,segment_array,number_of_segments)
char			*bin_mask;
Segment			**segment_array;
int32_t 		*number_of_segments;
{

  /* Local variables */
  int32_t	byte;
  int			bit;
  int32_t	bin_number;
  int			side;
  int32_t	adj_bin_number;
  int32_t	segment_count = 0;
  Segment		*segment = NULL;
  int			endpoint;
  double		lat,lon;

  /* Function declarations */
  int32_t	adjacent_bin_number();
  int			locate_quadsphere_bin();

  for(byte = 0; byte < QS_MASK_SIZE; byte++){	/* For every byte in the mask */
    for(bit = 0; bit < 8; bit++){		/* For every bit in each byte */
      if(bin_mask[byte] & 1<<bit){		/* If the bin is selected */

        bin_number = byte * 8  +  bit;		/* determine its bin number */

        /* QS_TOP and QS_LEFT should be defined as 0 and 3, respectively, in
         * quadsphere.h for the following to work properly.  I use the macros
         * instead of the numbers themselves in order to make the function of
         * the code clearer.
         *                          QS_TOP
         *                       .-----------.
         *                       |           |
         *                       |   quad-   |
         *               QS_LEFT |  sphere   | QS_RIGHT
         *                       |    bin    |
         *                       |           |
         *                       `-----------'
         *                         QS_BOTTOM
         */

        /* Move clockwise around the bin starting at the top. */
        for(side = QS_TOP; side <= QS_LEFT; side++){

          /* Determine the number of the adjacent bin. */
          adj_bin_number =
          adjacent_bin_number(bin_number,QS_BIN_LEVEL,side);

          if(adj_bin_number == -1){
            fprintf(stderr,"-W- %s line %d: ",__FILE__,__LINE__);
            fprintf(stderr,"adjacent_bin_number(%d,%d,%d) failed.\n",
            bin_number,QS_BIN_LEVEL,side);
            return(ADJ_BIN_NUM_ERROR);
          }

          /* If adjacent bin is not selected, then we are
           * at a boundary and should draw a line segment. */
          if( !(bin_mask[adj_bin_number/8] & 1 << (adj_bin_number & 7) ) ){

            /* Increase amount of memory allocated for stored line segments. */
            segment_count++;
            segment =
            (Segment *)realloc(segment,segment_count*sizeof(Segment));
            if(segment == NULL){
              fprintf(stderr,"-W- %s line %d: Memory allocation failed.\n",
              __FILE__,__LINE__);
              return(REALLOC_ERROR);
            }

            for(endpoint = 0; endpoint < 2; endpoint++){

              /* Make use of the fact that the leading ends (in a clockwise
               * sense) of each side of a quadsphere bin are defined with the
               * same value as the side itself (see quadsphere.h). */
              if(!
              locate_quadsphere_bin
              (bin_number,QS_BIN_LEVEL,((side + endpoint) & 3),&lat,&lon)
              ){
                fprintf(stderr,"-W- %s line %d: ",__FILE__,__LINE__);
                fprintf(stderr,
                "locate_quadsphere_bin(%d,%d,%d,&lat,&lon) failed.\n",
                bin_number,QS_BIN_LEVEL,((side + endpoint) & 3) );
                return(LOCATE_BIN_ERROR);
              }

              /* Store an endpoint of the line segment. */
              segment[segment_count-1][endpoint].lat  = (float)lat;
              segment[segment_count-1][endpoint].lon = (float)lon;

            }
          }
        }
      }
    }
  }

  /* There should be an even number of segments and that number should
   * be at least four since it takes four segments to surround one bin. */
/*
  if(segment_count & 1  ||  segment_count < 4)
    return(LOGIC_ERROR);
*/

  *segment_array      = segment;
  *number_of_segments = segment_count;

  return(SUCCESS);

}/*end generate_segment_array()*/









/*------------------------------------------------------------------------------

   Function: generate_linked_lists

   Returns type: int

   Description:	This function assembles the input array of quadsphere bin
		side specifications into one or more singly-linked lists
		of earth coordinates.  Each linked list represents the
		perimeter of one contiguous cluster of selected bins.

		This function returns one of the status values defined
		at the beginning of this file.

   Parameters: (in calling order)
      Type		Name			I/O	Description
      ----		----			---	-----------
      Segment *		segment_array		In	Array of pairs of
							earth-coordinate pairs
      long		number_of_segments	In	Number of elements in
							segment_array
      OutlineList *	pgons			Out	One or more spherical
							polygons

   Modification history:
      Programmer	Date		Description of change
      ----------	----		---------------------
      Norman Kuring	23-Feb-1993	Original development
      Norman Kuring	11-Mar-2004	Adapt to non-X-callback use

------------------------------------------------------------------------------*/

int
generate_linked_lists
(segment_array,number_of_segments,pgons)
Segment			*segment_array;
int32_t			number_of_segments;
OutlineList		*pgons;
{

  int32_t	first_segment;
  int32_t	number_of_unattached_segments;
  int32_t	segment_index;
  int			number_of_polygons;
  Outline		*polygon;
  CoordinateList	*vertex;
  int32_t	number_of_vertices;
  int			endpoint;

  /****************************************************************************/
  /* This loop generates singly-linked lists of earth coordinates for subsets
   * of segments in the segment array which form closed polygons when assembled
   * together.  Each pass through the loop creates another linked list.
   * The entry points into these linked lists are stored in the array,
   * array_of_lists.
   */
  for(
  first_segment = 0,
  number_of_polygons = 0, polygon = NULL,
  number_of_unattached_segments = number_of_segments;
  number_of_unattached_segments > 3;	/* Min. # of segments is 4 for 1 bin. */
  number_of_polygons++
  ){

    polygon = (Outline *)
    realloc(
    polygon,
    (number_of_polygons + 1) * sizeof(Outline));
    if(polygon == NULL){
      fprintf(stderr,"-W- %s line %d: Memory allocation failed.\n",
      __FILE__,__LINE__);
      return(REALLOC_ERROR);
    }

    polygon[number_of_polygons].vrtx =
    (CoordinateList *)malloc(sizeof(CoordinateList));
    if(polygon[number_of_polygons].vrtx == NULL){
      fprintf(stderr,"-W- %s line %d: Memory allocation failed.\n",
      __FILE__,__LINE__);
      return(MALLOC_ERROR);
    }

    vertex = polygon[number_of_polygons].vrtx;

    vertex->lat = segment_array[first_segment][0].lat;
    vertex->lon = segment_array[first_segment][0].lon;

    /* Allocate enough memory for another vertex and connect it. */
    vertex->next = (CoordinateList *)malloc(sizeof(CoordinateList));
    if(vertex->next == NULL){
      fprintf(stderr,"-W- %s line %d: Memory allocation failed.\n",
      __FILE__,__LINE__);
      return(MALLOC_ERROR);
    }

    vertex = vertex->next;

    vertex->lat = segment_array[first_segment][1].lat;
    vertex->lon = segment_array[first_segment][1].lon;
    vertex->next      = NULL;

    /* Now that we have stored this segment in a linked list,
     * we can erase it from the segment array. */
    for(endpoint = 0; endpoint < 2; endpoint++){
      segment_array[first_segment][endpoint].lat = 999;
      segment_array[first_segment][endpoint].lon = 999;
    }

    /* Advance the first-segment-index to the next unerased segment. */
    while(segment_array[++first_segment][0].lat == 999)
      ;

    /* Reflect the fact that we have one less segment to choose from now. */
    number_of_unattached_segments--;

    /**************************************************************************/
    /* This loop continues to assemble line segments
     * until the segments form a closed polygon.	*/
    for(
    number_of_vertices = 2;
    vertex->next == NULL;
    number_of_vertices++
    ){

      /************************************************************************/
      /* This loop searches the array of segments to
       * find one that attaches to the current vertex. */
      for(
      segment_index = first_segment;
      segment_index < number_of_segments;
      segment_index++
      ){

        /* Check each endpoint of the segment. */
        for(endpoint = 0; endpoint < 2; endpoint++)
          if(
          segment_array[segment_index][endpoint].lat == vertex->lat
          &&
          segment_array[segment_index][endpoint].lon == vertex->lon
          )
            break;
        if(endpoint < 2)
          break;		/* Matching segment endpoint found */

      }/***********************************************************************/

      if(segment_index == number_of_segments){	/* This should not happen. */
        fprintf(stderr,"-W- %s line %d: Program logic error.\n",
        __FILE__,__LINE__);
        return(LOGIC_ERROR);
      }

      else if(		/* If other end of segment matches initial vertex */
      segment_array[segment_index][1 - endpoint].lat	==
      polygon[number_of_polygons].vrtx->lat		&&
      segment_array[segment_index][1 - endpoint].lon	==
      polygon[number_of_polygons].vrtx->lon
      ){
        /* Close the polygon. */
        vertex->next = polygon[number_of_polygons].vrtx;
        /* Store the number of vertices contained in the list. */
        polygon[number_of_polygons].n = number_of_vertices;
      }

      else{

        /* Allocate enough memory for another vertex and connect it. */
        vertex->next = (CoordinateList *)malloc(sizeof(CoordinateList));
        if(vertex->next == NULL){
          fprintf(stderr,"-W- %s line %d: Memory allocation failed.\n",
          __FILE__,__LINE__);
          return(MALLOC_ERROR);
        }

        /* Move to the new vertex */
        vertex = vertex->next;

        /* Store coordinates for this vertex. */
        vertex->lat = segment_array[segment_index][1-endpoint].lat;
        vertex->lon = segment_array[segment_index][1-endpoint].lon;

        /* This vertex does not point anywhere yet. */
        vertex->next      = NULL;

      }

      /* Now that we have stored this segment in a linked list,
       * we can erase it from the segment array. */
      for(endpoint = 0; endpoint < 2; endpoint++){
        segment_array[segment_index][endpoint].lat = 999;
        segment_array[segment_index][endpoint].lon = 999;
      }

      /* If the first segment of the array has been erased then advance
       * the first-segment-index to the next unerased segment. */
      if(segment_index == first_segment)
        while(segment_array[++first_segment][0].lat == 999)
          ;

      /* Reflect the fact that we have one less segment to choose from now. */
      number_of_unattached_segments--;

    }/*************************************************************************/
  }/***************************************************************************/

  /* Make sure that we have connected all of the bin sides to each other. */
  if(number_of_unattached_segments){
    fprintf(stderr,"-W- %s line %d: Program logic error.\n",__FILE__,__LINE__);
    return(LOGIC_ERROR);
  }

  pgons->n    = number_of_polygons;
  pgons->pgon = polygon;

  return(SUCCESS);

}/*end generate_linked_lists()*/




/*------------------------------------------------------------------------------

   Function: find_geographical_extrema

   Returns type: int

   Description:	This function finds the smallest rectangle on a cylindrical
		equirectangular projection of the world that will completely
		contain all of the selected bins (in other words, all of
		those bins whose corresponding bit in the bin_mask portion
		of the AreaOfInterest structure is set) for the specified
		area of interest.  The AreaOfInterest structure is updated
		with the coordinate information.

		This function returns one of the status values listed at the
		beginning of this file.


   Parameters: (in calling order)
      Type		Name			I/O	Description
      ----		----			---	-----------
      AreaOfInterest *	area_of_interest	In/Out	See spatial_info.h

   Modification history:
      Programmer	Date		Description of change
      ----------	----		---------------------
      Norman Kuring	03-Mar-1993	Original development
      Norman Kuring	11-Mar-2004	Adapt to non-X-callback use
      Norman Kuring	31-Mar-2004	Compute a central lat/lon for the AOI

------------------------------------------------------------------------------*/

int
find_geographical_extrema(area_of_interest)
AreaOfInterest		*area_of_interest;
{

  /* Local variables */
  int			polygon_index;
  int			longitude_extrema_have_not_been_determined_yet = 1;
  int32_t		vertex_counter;
  CoordinateList	*vertex;
  float			westernmost_longitude;
  float			easternmost_longitude;
  int			longitude_multiplier;
  float			adjusted_longitude;
  float			minimum_adjusted_longitude;
  float			maximum_adjusted_longitude;
  float			difference[4];
  int			index;
  double		magn,cen[3] = {0,0,0};

  /* Function declaration */
  float			longitude_difference();


  /* Initialize this area's latitude extrema. */
  area_of_interest->s      =  90.0;
  area_of_interest->n      = -90.0;

  /* If north-pole bin is set, then adjust the coordinates accordingly. */
  if(
  area_of_interest->bmsk[QS_LEVEL_7_NORTH_POLE_BIN_NUMBER/8]
  &  1 << (QS_LEVEL_7_NORTH_POLE_BIN_NUMBER & 7)
  ){
    area_of_interest->n =   90.0;
    area_of_interest->w = -180.0;
    area_of_interest->e =  180.0;
    longitude_extrema_have_not_been_determined_yet = 0;
  }

  /* If south-pole bin is set, then adjust the minimum latitude accordingly. */
  if(
  area_of_interest->bmsk[QS_LEVEL_7_SOUTH_POLE_BIN_NUMBER/8]
  &  1 << (QS_LEVEL_7_SOUTH_POLE_BIN_NUMBER & 7)
  ){
    area_of_interest->s =  -90.0;
    area_of_interest->w = -180.0;
    area_of_interest->e =  180.0;
    longitude_extrema_have_not_been_determined_yet = 0;
  }

  for(
  polygon_index = 0;
  polygon_index < area_of_interest->pgons.n;
  polygon_index++){

    vertex = area_of_interest->pgons.pgon[polygon_index].vrtx;

    if(longitude_extrema_have_not_been_determined_yet){
      minimum_adjusted_longitude = vertex->lon;
      maximum_adjusted_longitude = vertex->lon;
      longitude_multiplier = 0;
    }

    for(
    vertex_counter = 0;
    vertex_counter <
    area_of_interest->pgons.pgon[polygon_index].n;
    vertex_counter++,
    vertex = vertex->next
    ){
      double	coslat;

      if(vertex == NULL || vertex->next == NULL){
        fprintf(stderr,"-W- %s line %d: ",__FILE__,__LINE__);
        fprintf(stderr,"Inconsistent data in area-of-interest outline.\n");
        return(INCONSISTENT_DATA_ERROR);
      }

      /*
      By adding boundary position vectors
      I can come up with a central position.
      */
      coslat = cos(vertex->next->lat * PI/180);
      cen[0] += coslat * cos(vertex->next->lon * PI/180);
      cen[1] += coslat * sin(vertex->next->lon * PI/180);
      cen[2] +=          sin(vertex->next->lat * PI/180);

      if(vertex->next->lat < area_of_interest->s)
        area_of_interest->s = vertex->next->lat;

      if(vertex->next->lat > area_of_interest->n)
        area_of_interest->n = vertex->next->lat;

      if(longitude_extrema_have_not_been_determined_yet){

        /* If dateline crossed heading eastward */
        if( vertex->next->lon < -90
        &&  vertex->lon       >  90 )
          longitude_multiplier++;

        /* If dateline crossed heading westward */
        else
        if( vertex->next->lon >  90
        &&  vertex->lon       < -90 )
          longitude_multiplier--;

        adjusted_longitude =
        vertex->next->lon + 360.0 * longitude_multiplier;

        if(adjusted_longitude < minimum_adjusted_longitude)
          minimum_adjusted_longitude = adjusted_longitude;

        if(adjusted_longitude > maximum_adjusted_longitude)
          maximum_adjusted_longitude = adjusted_longitude;

      }

    }

    if(longitude_extrema_have_not_been_determined_yet){

      /* Put the longitudes back within the range, (-180,180). */
      if( maximum_adjusted_longitude - minimum_adjusted_longitude  >=  360.0){
        westernmost_longitude = -180.0;
        easternmost_longitude =  180.0;
      }
      else{
        westernmost_longitude =
        minimum_adjusted_longitude +
        (minimum_adjusted_longitude < -180 ? 360 : 0);

        easternmost_longitude =
        maximum_adjusted_longitude -
        (maximum_adjusted_longitude >  180 ? 360 : 0);
      }

      if(polygon_index == 0){	/* Initial determination of longitude extrema */

        area_of_interest->w = westernmost_longitude;
        area_of_interest->e = easternmost_longitude;

      }
      else{			/* Determination of longitude extrema for
				 * the inclusion of subsequent polygons */

        /* When adding another polygon to the already determined rectangular
         * (in the cylindrical equirectangular sense) area of interest, I would
         * like to choose the smallest possible rectangle that still includes
         * all of the polygons encountered so far.  To do this I must allow the
         * rectangle to span the dateline if necessary.  So, there are four
         * different combinations of westernmost and easternmost longitudes
         * which could form the boundaries of the new rectangle.  Determine the
         * number of degrees traversed when moving from the west to the east
         * for each of the four cases.
         */
        difference[0] =
        longitude_difference(
        westernmost_longitude,			/* Boundaries of polygon */
        easternmost_longitude			/* to be added           */
        );
        difference[1] =
        longitude_difference(
        area_of_interest->w,			/* Boundaries of previously */
        area_of_interest->e			/* determined rectangle     */
        );
        difference[2] =
        longitude_difference(			/* Western boundary of   */
        area_of_interest->w,			/* rectangle and eastern */
        easternmost_longitude			/* boundary of polygon   */
        );
        difference[3] =
        longitude_difference(			/* Western boundary of   */
        westernmost_longitude,			/* polygon and eastern   */
        area_of_interest->e			/* boundary of rectangle */
        );

        /* From the four differences calculated above, pick the greatest value
         * among the first difference (index = 0), the second difference (index
         * = 1), and the minimum of the third and fourth differences (index = 2
         * and index = 3).
         */
        index = difference[3]     < difference[2] ? 3     : 2;
        index = difference[index] > difference[1] ? index : 1;
        index = difference[index] > difference[0] ? index : 0;

        /* The longitudes that correspond to the difference selected above
         * are the new boundaries for the rectangular area of interest. */
        switch(index){
          case 0:
            area_of_interest->w = westernmost_longitude;
            area_of_interest->e = easternmost_longitude;
            break;
          case 1:
            break;
          case 2:
            area_of_interest->e = easternmost_longitude;
            break;
          case 3:
            area_of_interest->w = westernmost_longitude;
            break;
        }
      }
    }
  }

  /* Convert central vector into a latitude and longitude. */
  magn = sqrt(cen[0]*cen[0] + cen[1]*cen[1] + cen[2]*cen[2]);
  cen[0] /= magn;
  cen[1] /= magn; 
  cen[2] /= magn;
  area_of_interest->clat =
  atan2(cen[2],sqrt(cen[0]*cen[0] + cen[1]*cen[1]))*180/PI;
  area_of_interest->clon = atan2(cen[1],cen[0])*180/PI;

  /* Check validity of output. */
  if(area_of_interest->s <  -90.0
  || area_of_interest->s >   90.0
  || area_of_interest->n <  -90.0
  || area_of_interest->n >   90.0
  || area_of_interest->w < -180.0
  || area_of_interest->w >  180.0
  || area_of_interest->e < -180.0
  || area_of_interest->e >  180.0){
    fprintf(stderr,"-W- %s line %d: ",__FILE__,__LINE__);
    fprintf(stderr,
    "Invalid extrema have been computed for an area of interest.\n");
    return(COORDINATE_ERROR);
  }

  return(SUCCESS);

}/*end find_geographical_extrema()*/
