/*------------------------------------------------------------------------------

   Function: locate_quadsphere_bin

   Returns type: int

   Description:

        ROUTINE TO COMPUTE LATITUDE AND LONGITUDE OF BIN CENTER
        FROM QUAD-SPHERE BIN NUMBER 

        Program logic generated by:
        Frederick S. Patt / General Sciences Corporation
        July 16, 1992


Quad Sphere

The Earth's surface is projected onto the faces of an inscribed cube using
a curvilinear projection which preserves area.  The sphere is divided into
six equal areas which correspond to the faces of the cube.  The vertices of
the cube correspond to the cartesian coordinates defined by |x|=|y|=|z| on
the unit sphere.  The cube is oriented with one face normal to the North
Pole and one face centered on the Greenwich meridian.  

The faces of the cube are divided into square bins, where the number of
bins along each edge is a power of 2, selected to produce the desired bin
size.  Thus the number of bins on each face is 2**(2*N), where N is the
binning level, and the total number of bins is 6*2**(2*N).  For example, a
level of 10 gives 1024x1024 bins on each face and 6291456 (6*2**20) total
bins, and on the Earth's surface the bins are 81 km**2 in size. 

The bins are numbered serially, and the bin numbers are determined as
follows.  The total number of bits required for the bin numbers at level N
is 2*N+3, where the 3 MSBs are used for the face numbers and the remaining
bits are used to number the bins within each face.  The faces are numbered
0-5 with 0 being the North face, 1 through 4 being equatorial with 1
corresponding to Greenwich, and 5 being South.  Thus at level 10, face 0
has bin numbers 0-1048577, face 1 has numbers 1048576-2097151, etc.  

Within each face the bins are numbered serially from one corner (the
convention is to start at the "lower left") to the opposite corner, with
the ordering such that each pair of bits corresponds to a level of bin
resolution.  This ordering in effect is a two-dimensional binary tree,
which is referred to as the quad-tree  The conversion between bin numbers
and coordinates is straightforward.  The maximum practical bin level is 14,
which uses 31 bits in a 32-bit integer and results in a bin size of 0.32
km**2. 

The quad sphere limits the allowable bin sizes by requiring the bin 
dimenension of the cube face to be a power of 2, due to the binary indexing 
scheme.  The advantage is that the bin numbering allows the resolution to
be changed (by factors of two) simply by adding or deleting LSBs.  This is
particularly useful if it is desired to increase the bin size, for example
to compare maps with different resolutions.  This is performed simply by
dividing the bin numbers by four for each factor of two increase in bin
size. 

The quad sphere faces can be displayed individually without remapping, with 
moderate distortion in the corners of the faces.

The quad sphere has an advantage in that the binary numbering scheme allows
contiguous subsets of bins (either whole faces or quadrants of faces) to be
accessed by simply specifying a range of bin numbers.


        High-Level Logic:

    1.  Determine face number from bin number using level number
    2.  Compute face coordinates (u,v) of bin center from bin number 
    3.  Compute cartesian coordinates from face coordinates 
    4.  Compute LAT and LON from cartesian coordinates 

        Detailed Logic:

    1.  DETERMINE FACE NUMBER FROM BIN NUMBER USING LEVEL NUMBER

        Determine face number as high order bits of bin number

        IFACE = BIN/(2**(2*LEV))
        
        Check for invalid bin number 
        IF ((IFACE.GT.5).OR.(BIN.LT.0) THEN
           Signal error condition 
           RETURN
        END IF

    2.  COMPUTE FACE COORDINATES (U,V) FROM BIN NUMBER 

        Parse bin number into U and V indices

        ITEMP = BIN - IFACE*2**(2*LEV)

        IU = 0
        IV = 0
        IN = 1
        DOWHILE (ITEMP.GT.0)
            Strip off U bit
            IT = MOD(ITEMP,2)
            IU = IU + IT*IN
            ITEMP = ITEMP/2

            Strip off V bit             
            IT = MOD(ITEMP,2)
            IV = IV + IT*IN
            ITEMP = ITEMP/2
        
            Increase bit value by a factor of 2
            IN = IN*2
        END DO

        Convert (U,V) from range of (0,2**LEV-1) to (-1.,+1.) 
        U = (2.*IU+1)/(2**LEV) - 1.
        V = (2.*IV+1)/(2**LEV) - 1.
        

    3.  COMPUTE CARTESIAN COORDINATES FROM FACE COORDINATES 
            PROJECTION EQUATIONS ASSUME KNOWLEDGE OF LARGER OF (U,V)

        Project face coordinates to cartesian vector
        IF (ABS(U).GE.ABS(V)) THEN
            COMPUTE (Q,R,S) FROM (U,V) 
            SOR = SIN((PI*V)/(12.*ABS(U)))/(COS((PI*V)/(12.*ABS(U)))-SQRT(0.5))
            Q = 1. - (1.-1./SQRT(2.+(SOR)**2))*U**2
            R = SQRT((1.-Q**2)/(1.+SOR**2))
            S = SOR*R
            IF (U.LT.0) R = -R
        ELSE 
            COMPUTE (Q,S,R) FROM (V,U) 
            ROS = SIN((PI*U)/(12.*ABS(V)))/(COS((PI*U)/(12.*ABS(V)))-SQRT(0.5))
            Q = 1. - (1.-1./SQRT(2.+(ROS)**2))*V**2
            S = SQRT((1.-Q**2)/(1.+ROS**2))
            R = ROS*S
            IF (V.LT.0) S = -S
        END IF

        Transform vector to reference frame according to face number
            IF (IFACE.EQ.0) THEN 
                X = -S
                Y = R
                Z = Q
            ELSE IF (IFACE.EQ.1) THEN
                X = Q
                Y = R
                Z = S
            ELSE IF (IFACE.EQ.2) THEN 
                X = -R
                Y = Q
                Z = S
            ELSE IF (IFACE.EQ.3) THEN
                X = -Q
                Y = -R
                Z = S
            ELSE IF (IFACE.EQ.4) THEN
                X = R
                Y = -Q
                Z = S
            ELSE 
                X = S
                Y = R
                Z = -Q
            END IF


    4.  COMPUTE LAT AND LON FROM CARTESIAN COORDINATES 

        LAT = ASIN(Z)
        LON = ATAN2(Y,X)

        END DETAILED LOGIC

	This function returns 1 upon successful completion
	and 0 in case of error.


   Parameters: (in calling order)
      Type		Name		I/O	Description
      ----		----		---	-----------
      unsigned long	bin_number	In	Quadsphere bin number
      int		level		In	Quadsphere level (see descrip.)
      int		desired_coordinates
					In	Do we want a corner or center?
      double *		latitude	Out	(degrees)
      double *		longitude	Out	(degrees)

   Modification history:
      Programmer	Date		Description of change
      ----------	----		---------------------
      Norman Kuring	20-Oct-1992	Original development (find center only)
      Norman Kuring	16-Feb-1993	Locate bin corner or center

------------------------------------------------------------------------------*/

#include <math.h>
#include "quadsphere.h"

#define		SUCCESS		1
#define		ERROR		0

#define PI	3.14159265358979323846
#define SQRT1_2	0.70710678118654752440

int
locate_quadsphere_bin
(bin_number,level,desired_coordinates,latitude,longitude)
unsigned long	bin_number;
int		level;
int		desired_coordinates;
double		*latitude;
double		*longitude;
{

  int		cube_face;
  int		iu,iv,in;
  int		iu_offset,iv_offset;
  double	u,v;
  double	sor,ros,q,r,s;
  double	x,y,z;

  if(level > 14 || level < 1)	/* Binning level limited to    */
    return(ERROR);		/* 14 by size of long integer. */

  switch(desired_coordinates){
    case QS_CENTER:
      iu_offset = 1; iv_offset = 1;
      break;
    case QS_TOP_LEFT:
      iu_offset = 0; iv_offset = 2;
      break;
    case QS_TOP_RIGHT:
      iu_offset = 2; iv_offset = 2;
      break;
    case QS_BOTTOM_LEFT:
      iu_offset = 0; iv_offset = 0;
      break;
    case QS_BOTTOM_RIGHT:
      iu_offset = 2; iv_offset = 0;
      break;
    case QS_TOP_CENTER:
      iu_offset = 1; iv_offset = 2;
      break;
    case QS_BOTTOM_CENTER:
      iu_offset = 1; iv_offset = 0;
      break;
    case QS_LEFT_CENTER:
      iu_offset = 0; iv_offset = 1;
      break;
    case QS_RIGHT_CENTER:
      iu_offset = 2; iv_offset = 1;
      break;
    default:
      return(ERROR);
  }
      

  cube_face = bin_number/(1 << 2*level);

  if(cube_face > 5 || bin_number < 0)
    return(ERROR);

  bin_number -= cube_face * (1 << 2*level);

  for(iu = 0, iv = 0, in = 1; bin_number; in *= 2){

    iu += (bin_number & 1) * in;
    bin_number >>= 1;

    iv += (bin_number & 1) * in;
    bin_number >>= 1;

  }

  u = ( (double)(2 * iu + iu_offset) ) / (1 << level) - 1.0;
  v = ( (double)(2 * iv + iv_offset) ) / (1 << level) - 1.0;

  if(fabs(u) >= fabs(v)){

    if(u == 0.0 && v == 0.0){	/* Requested coordinates are at */
      q = 1.0;			/* the center of a cube face.   */
      r = 0.0;
      s = 0.0;
    }
    else{
      sor = sin((PI*v)/(12*fabs(u))) / (cos((PI*v)/(12*fabs(u)))-SQRT1_2);
      q = 1 - (1 - 1/sqrt(2 + sor*sor)) * u*u;
      r = sqrt((1 - q*q)/(1 + sor*sor));
      s = sor * r;
      if(u < 0)
        r = -r;
    }

  }
  else{

    ros = sin((PI*u)/(12*fabs(v))) / (cos((PI*u)/(12*fabs(v))) - SQRT1_2);
    q = 1 - (1 - 1/sqrt(2 + ros*ros)) * v*v;
    s = sqrt((1 - q*q)/(1 + ros*ros));
    r = ros * s;
    if(v < 0)
      s = -s;

  }

  switch(cube_face){

    case 0:	x = -s;   y =  r;   z =  q;	break;
    case 1:	x =  q;   y =  r;   z =  s;	break;
    case 2:	x = -r;   y =  q;   z =  s;	break;
    case 3:	x = -q;   y = -r;   z =  s;	break;
    case 4:	x =  r;   y = -q;   z =  s;	break;
    case 5:	x =  s;   y =  r;   z = -q;	break;
    default:
      return(ERROR);

  }

  *latitude  = asin(z)    * 180/PI;		/* Give outputs in degrees */
  *longitude = atan2(y,x) * 180/PI;
  if(*longitude == 180)
    *longitude = -180.0;			/* Be consistent */

  return(SUCCESS);

}/*end locate_quadsphere_bin()*/
