/*
Norman Kuring		22-Mar-2004	Original development
Norman Kuring		31-Mar-2004	Add central lat/lon to mode 1 output
Norman Kuring		21-Jul-2005	Fixed bug in mask generation
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include "spatial_info.h"

#define USAGE "\
Usage:\n\
%s mode [qbinlistfile]\n\
where mode         is either 1 (output geographical extrema\n\
                                and central coordinate)\n\
                      or     2 (output geographical outlines)\n\
and   qbinlistfile is the path to a file containing quadsphere (level 6)\n\
                      bin numbers one per line.  If qbinlistfile is not\n\
                      specified, then the bin numbers will be read from\n\
                      standard input.\n\
"

#define MAXSTRINGSIZE		1024

int generate_area_outlines(ListOfAreas loa);

int main(int argc, char *argv[]){

  int		mode;
  FILE		*fh;
  char		string[MAXSTRINGSIZE];
  ListOfAreas	loa;
  int		status;

  if(argc < 2){
    fprintf(stderr,USAGE,argv[0]);
    return(EXIT_FAILURE);
  }

  mode = atoi(argv[1]);
  if(mode != 1 && mode != 2){
    fprintf(stderr,USAGE,argv[0]);
    return(EXIT_FAILURE);
  }

  if(argc == 2){
    fh = stdin;
  }
  else{
    if((fh = fopen(argv[2],"r")) == NULL){
      fprintf(stderr,"-E- %s line %d: Could not open file, \"%s\". ",
      __FILE__,__LINE__,argv[2]);
      perror("");
      return(EXIT_FAILURE);
    }
  }

  /* Initialize one area of interest. */
  loa.area = (AreaOfInterest *)calloc(1,sizeof(AreaOfInterest));
  if(loa.area == NULL){
    fprintf(stderr,"-E- %s line %d: Memory allocation failed.\n",
    __FILE__,__LINE__);
    return(EXIT_FAILURE);
  }
  loa.n = 1;

  /* Read quadsphere bin numbers and set the appropriate bits in the mask. */
  while(fgets(string,MAXSTRINGSIZE,fh)){
    char	*check;
    int32_t	bin;

    bin = strtol(string,&check,10);
    if(check == string){
      /* Conversion failed.  Skip this string. */
      string[strcspn(string,"\n")] = 0;
      fprintf(stderr,"-W- %s line %d: Could not convert \"%s\" ",
      __FILE__,__LINE__,string);
      fprintf(stderr," to an integer.\n");
      continue;
    }
    if(bin < 0 || bin > 24575){
      /*
      A quadsphere level-6 bin number must be an
      integer in the range 0 to 24575, inclusive.
      */
      string[strcspn(string,"\n")] = 0;
      fprintf(stderr,
      "-W- %s line %d: bin number (%s) out of range [0,24575].\n",
      __FILE__,__LINE__,string);
      continue;
    }
    *( loa.area->bmsk + (bin >> 3) ) |= 1L << (bin & 7L);
  }

  /* Convert the bin mask to one or more area outlines. */
  status = generate_area_outlines(loa);
  if(status != 0){
    fprintf(stderr,"-E- %s line %d: generate_area_outlines() failed.",
    __FILE__,__LINE__);
    fprintf(stderr," status = %d\n",status);
    return(EXIT_FAILURE);
  }

  fclose(fh);

  if(mode == 1){
    printf("%f %f %f %f %f %f\n",
    loa.area->n,loa.area->s,loa.area->w,loa.area->e,
    loa.area->clat         ,loa.area->clon          );
  }
  else{
    int		i;

    for(i=0; i<loa.area->pgons.n; i++){
      int		j;
      CoordinateList	*v;

      if(i > 0) printf(">\n");	/* Separate different polygons with ">". */

      v = loa.area->pgons.pgon[i].vrtx;

      for(j=0; j<loa.area->pgons.pgon[i].n; j++){
        printf("%8.4f %9.4f\n",v->lat,v->lon);
        v = v->next;
      }
    }
  }

  free(loa.area);

  return(EXIT_SUCCESS);
}
